/*
 * Single modular addition in base 2 Kuperberg's algorithm simulator
 */


use std::mem::swap;
use std::env;

use std::cmp::min;
extern crate rand;
use rand::Rng;


// List of functions our integer type must have
// Currently implemented for u64 and [u64 ;2], allowing simulations up to 128 bits
trait Element : Default + Copy {

  // Arithmetic : Addition, substraction
  fn add(self, Self) -> Self;
  fn sub(self, Self) -> Self;
  
  // Getter : get the boolean 
  fn get_bool(&self, pos : usize) -> bool;
  
  //Gives the number of colliding bits between two elements
  fn common(&self, &Self) -> usize;
  
  // Check if an element is 2^{n-1}
  fn is_ok(&self, usize) -> bool;
  
  // Create random elements from a RNG
  fn make_elt(n : usize, mut g : &mut Rng) -> Self;
  
  // If the number is not of the form (.*)01(0*), negate it.
  // This allows to check only for collisions, as the case a+b will always be the bad one.
  fn normalize(&mut self);
}

impl Element for [u64 ; 2] {

  fn add(self, e2 : Self) -> Self{
    let (r0, b0) = self[0].overflowing_add(e2[0]);
    [ r0, self[1].wrapping_add(e2[1]) + if b0 {1} else {0}]
  }
  
  fn sub(self, e2 : Self) -> Self{
    let (r0, b0) = self[0].overflowing_sub(e2[0]);
    let r1 = self[1].wrapping_sub(e2[1]).wrapping_sub(if b0 {1} else {0});
    [ r0, r1]
  }
  
  fn get_bool(&self, pos : usize) -> bool{
    (if pos >= 64 {
        self[1] & (1 << (pos - 64))
      } else {
      self[0] & (1 << pos) } ) != 0
  }
  
  fn common(&self, e2 : &Self) -> usize{
    let mut z = self[0].wrapping_sub(e2[0]).trailing_zeros();
    if z == 64 { z+= self[1].wrapping_sub(e2[1]).trailing_zeros()}
    z as usize
  }
  
  fn is_ok(&self, n : usize) -> bool {
    match n {
      n if n <= 63 => {self[0] % (1<<n) == (1 << (n - 1))},
      64 => {self[0] == 1 << 63},
      128 => {(self[0] == 0) & (self[1] == 1 << 63)},
      n => {(self[0] == 0) & (self[1] % (1 << (n - 64)) == (1 << (n - 65)))},
    }
  }
  
  fn make_elt(n : usize, mut g : &mut Rng) -> Self {
    [ g.next_u64(), g.next_u64()] 
  }
  
  fn normalize(&mut self) {
    match self[0].trailing_zeros() {
        64 => { match self[1].trailing_zeros() {
                 63|64 => {}
                 z  => {if self[1] & (1 << (z+1)) != 0 {self[1]=self[1].wrapping_neg()}}
              }
            }
      63 => {if self[1] & 1 != 0 {self[1]= self[1].wrapping_neg()}}
      z  => {if self[0] & (1 << (z+1)) != 0 { *self = [0,0].sub(*self)}}
    }
  }
}

impl Element for u64 {

  fn add(self, e2 : Self) -> Self{
    self.wrapping_add(e2)
  }

  fn sub(self, e2 : Self) -> Self{
    self.wrapping_sub(e2)
  }

  fn get_bool(&self, pos : usize) -> bool{
    self & (1 << pos) != 0
  }

  fn common(&self, e2 : &Self) -> usize{
    self.wrapping_sub(*e2).trailing_zeros() as usize
  }

  fn is_ok(&self, n : usize) -> bool {
    match n {
      64 => {*self == 1 << 63},
      n  => {self % (1<<n) == (1<<(n-1))},
    }
  }

  fn make_elt(n : usize, mut g : &mut Rng) -> Self {
    g.next_u64() >> (64 - n)
  }
  
  fn normalize(&mut self){
    match self.trailing_zeros() {
      63|64 => {}
      z => { if *self & (1 << (z+1)) != 0 { *self = 0.sub(*self)} }
    }
  }
}


#[derive(Debug)]
enum Tree<T : Element>{
  Empty,
  Leaf { nb : i32, leaf : T },
  Node { len : i32, best : bool, lchild : Box<Tree<T>>, rchild : Box<Tree<T>>, nb : usize }
}

use Tree::*;


//Create the needed nodes for a new element
fn new_node<T : Element>(c: usize, t : &mut Tree<T>, r : T, pos : usize, nb : i32){
  let b = r.get_bool(c);
  let common  = r.clone();
  let new_leaf = Leaf {nb : 1, leaf : r};
  let nbs = nb as usize +1;
  let mut old_leaf = Empty;
  swap(t, &mut old_leaf);
  let (l,r) = if b {(old_leaf,new_leaf)} else {(new_leaf,old_leaf)};
  let mut tree = Node { len : 0, best : false, lchild : Box::new(l), rchild : Box::new(r) , nb : nbs};
  for i in 1..(c + 1 - pos) {
    let e = common.get_bool(c - i);
    let (lchild, rchild) = if e {(Box::new(Empty), Box::new(tree))} else {(Box::new(tree), Box::new(Empty))};
    tree = Node { len : i as i32, best : e, lchild : lchild, rchild : rchild, nb : nbs };
  }
  swap(t, &mut tree)
}

// Gives the length of the best collision in the tree, -1 means no collision
fn length<T : Element>(t : & Tree<T>) -> i32{
  match *t {
    Empty | Leaf { .. } => -1,
    Node { len : t, .. } => t
  }
}

// Insert recursively a new element in the tree
fn insert<T : Element>(e: T, tree: &mut Tree<T>, pos: usize, n: usize) -> i32{
    match tree {
        t @ &mut Empty => { *t = Leaf{ nb : 1, leaf : e }; -1 },
        t @ &mut Leaf{ .. } => { 
          let (c, nb) = match t {
            &mut Leaf{ ref mut leaf, ref mut nb } => { let g = leaf.common(&e); if g >= n {*nb+=1; return -1}; (g, *nb) },
            _ => unreachable!()
          };
          new_node(c, t , e, pos, nb); 
          (c - pos) as i32
	}
        &mut Tree::Node { ref mut len, ref mut best, ref mut lchild, ref mut rchild, ref mut nb } => {
            *nb+=1;
            if e.get_bool(pos) {
                let l2 = insert(e, rchild, pos+1, n);
            	if l2 >= *len{
                	*best = true;
                	*len = l2+1;
            	}
            } else {
                let l2 = insert(e, lchild, pos+1, n);
            	if l2 >= *len{
            	    *best = false;
            	    *len = l2+1;
            	}
            }
            *len
        }
    }
}


// Normalize the element, then insert it
fn canonical_insert<T : Element>(mut e: T, tree: &mut Tree<T>, pos: usize, n: usize){
    e.normalize();
    insert(e, tree, pos, n);
}


// Simplify the tree if necessary after popping an element
fn sanitize<T : Element>(t : &mut Tree<T>){
 let mut garbage = Empty;
 match t {
   &mut Leaf { nb : 0, .. } => { },
   &mut Node { ref mut lchild, ref mut rchild, .. } => {
     match (lchild.as_mut(), rchild.as_mut()) {
       (&mut Empty, &mut Empty) => { }
       (&mut Empty, r @ &mut Leaf { .. }) => { swap(&mut garbage, r) },
       (l @ &mut Leaf { .. }, &mut Empty) => { swap(&mut garbage, l) },
       _ => { return }
       }
   },
   _ => { return }
 };
 swap(t, &mut garbage)
}

fn pop_single<T : Element>(t : & Tree<T>) -> T {
  match t {
  &Empty => {T::default()},
  &Leaf{ref leaf,..} => {leaf.clone()},
  &Node {best, ref lchild, ref rchild, ..} =>{pop_single(if best {rchild.as_ref()} else {lchild.as_ref()})}
  }
}

// Pop the two elements with the best potential collision
fn pop_best<T : Element>(t : &mut Tree<T>) -> (T, T){
  let (e1,e2) = match t {
    &mut Empty | &mut Leaf { nb : 1, .. } => { panic!("Pop'd in an empty tree") },
    &mut Leaf { ref mut nb, leaf : ref l } => { *nb-=2; (l.clone(), l.clone()) },
    &mut Node { ref mut len, ref mut best, ref mut lchild, ref mut rchild, ref mut nb } => {
      *nb-=2;
      let (e1, e2, tosan) = match (lchild.as_mut(), rchild.as_mut()) {
        (&mut Leaf { nb : ref mut nbl, leaf : ref l }, &mut Leaf { nb : ref mut nbr, leaf : ref r }) =>
          { *nbl-=1; *nbr-=1; (l.clone(), r.clone(), true) },
        (lchild @ _, rchild @ _) => {
          let (e1, e2) = if *best {pop_best(rchild)} else {pop_best(lchild)};
          let (lleft, lright) = (length(lchild), length(rchild));
          if lleft > lright {
            *len = lleft+1;
            *best = false;
          } else {
            *len = lright+1;
            *best = true;
          }
        (e1, e2, false)
        }
      };
     if tosan {sanitize(lchild); sanitize(rchild)};
     (e1, e2)
     }
  };
  sanitize(t);
  (e1,e2)
}


fn usage(s : &str){
 println!("Usage : {} BITS #SAMPLES #ITERATIONS", s); 
}



static mut state : u32 = 0;
static mut lstate : u32 = 0;

// Get a random boolean from a RNG
fn random_bool(g : &mut Rng) -> bool {
  unsafe {
    if lstate == 0 {
      lstate = 31;
      state = g.next_u32();
    } else {
      lstate-=1;
    }
    let b = state % 2 == 1;
    state/=2;
    b
  }
}

// Get the number of elements in a tree
fn num_elt<T : Element>(t : &Tree<T>) -> i32 {
  match t {
  &Empty => {0},
  &Leaf{ nb, .. } => {nb}
  &Node{nb, ..} => {nb as i32}
  }
}

// Gives the number of zeros in an element
fn zeros<T : Element>(e : &T, n: usize) -> usize {
  min(e.common(&Default::default()),n)
}

fn main(){
    //Argument parsing
    let args : Vec<String> = env::args().collect();
    if args.len() != 4 { return usage(&args[0]) };
    let (n, s, i) = match (args[1].parse(),args[2].parse(),args[3].parse()) {
      (Ok(n), Ok(s), Ok(i)) => {(n,s,i)},
      _ => { return usage(&args[0]) }
    };
    
    // For a cryptographic RNG : rand::thread_rng()
    let mut g = rand::weak_rng();
    
    //Main iteration loop
    let mut ok = 0i32;
    for _ in 0..i {
      let mut tree = Empty;

      let mut distro = vec![0 ; n+1];
      for _ in 0..s {
      
        // Choose here the type of the element
        let e : [ u64 ; 2 ] = Element::make_elt(n, &mut g);
	
	if e.is_ok(n) {distro[n-1]+=1} else {
          canonical_insert(e, &mut tree, 0, n);
        }
      }

      let mut pos = 1;
        loop {
	  // Get the best elements, combine them & iterate, make sure to keep a trail of uninteresting elements
          match tree {
            Empty => { break },
            Leaf { nb , ref leaf } if nb <= 2 => {distro[zeros(leaf, n)]+= nb;break },
            ref mut t @ Leaf{ .. } => {
              let (e1,e2) = pop_best(t);
              let e = if random_bool(&mut g) {e1.add(e2)} else {e1.sub(e2)};
              if e.is_ok(n) {distro[n-1]+=1} else {
	            canonical_insert(e, t, pos-1, n)}
	      }
	    Node { mut lchild , mut rchild, .. } => {
	      while num_elt(rchild.as_ref()) >=3 {
	        let (e1,e2) = pop_best(&mut rchild);
	        let e = if random_bool(&mut g) {e1.add(e2)} else {e1.sub(e2)};
		if e.is_ok(n) {distro[n-1]+=1} else {
	        canonical_insert(e, &mut lchild, pos, n) }
	      };
	      tree = *lchild;
	      distro[zeros(&pop_single(&rchild), n)] = num_elt(&rchild);
	      pos+=1;
	    }
          }
        }
      if distro.iter().take(n).all(|&x| x > 0) {ok+=1} else {}
      // Uncomment to print the index of the last bit we manage to get
      //println!("{}",distro.iter().enumerate().rev().skip(1).skip_while(|&(_,&v)| v > 0).next().unwrap().0 ) }
    }
    println!("{}",(ok as f64 / i as f64))
}
